import warnings
from typing import Any, ClassVar, Dict, List, Union


class Format:
    """
    An abstract base class for generating reference formats
    from CITATION.cff yaml-content.

    Constructed by passing a dict containing the yaml-processed
    content of the CITATION.cff file to the constructor.
    Required fields are checked for on instantiation.
    Missing optional fields will be set to None if not provided.

    Parameters
    ----------
    information : Dict[str, str]
        yaml-processed content of CITATION.cff.
    warn_on_not_used: bool, default = False
        If True, warn the user about the fields in the information
        input that are not required nor optional, so are ignored.

    Attributes
    ----------
    author_separator : str, default = ", "
        Separator to use between authors in the event the reference
        has more than one author provided.
    indent_character : str, default = ' ' * 4
        Character sequence to use for indentation.
    required : ClassVar[List[str]]
        Required fields for the entry to be created.
    optional : ClassVar[List[str]]
        Optional fields that the entry may possess.
    """

    indent_character: str = " " * 4
    author_separator: ClassVar[str] = ", "
    required: ClassVar[List[str]]
    optional: ClassVar[List[str]]

    # mypy type-hints
    authors: Union[str, Dict[str, str], List[Dict[str, str]]]

    @classmethod
    def entry_type(cls) -> str:
        """
        The reference format type; article, book, software, etc.
        This will be inherited by child classes and used to infer
        which format the citation.cff files are providing.
        """
        return cls.__name__.lower()

    def __init__(
        self,
        information: Dict[str, Any],
        warn_on_not_used: bool = False,
    ) -> None:
        """ """
        # So we don't delete information that we may need in other places
        # (C++ memory sharing rights plz Python)
        information = information.copy()

        # Add all the information we need
        for key, value in information.items():
            if key in self.required or key in self.optional:
                setattr(self, key.replace("-", "_"), value)
            elif warn_on_not_used:
                warnings.warn(
                    f"The key {key} is not used for entries of type "
                    f"{self.entry_type()}",
                    UserWarning,
                )

        # Check that all required information is populated
        for required_field in self.required:
            if not hasattr(self, required_field):
                raise KeyError(
                    f"Did not receive value for required key: {required_field}"
                )
        # Optional fields should be set to None so that checks against
        # them produce nothing and evaluate to False
        for optional_field in self.optional:
            if not hasattr(self, optional_field.replace("-", "_")):
                setattr(self, optional_field.replace("-", "_"), None)

        if hasattr(self, "authors"):
            self._prepare_authors_field()

        return

    def _prepare_authors_field(self) -> None:
        """
        The authors field may come in as a dict, or a list of dicts.
        This is rather inconvenient as we need it to be a string, so
        convert it with this function.

        Individual author fields in CITATION.cff are:
        - family-names
        - given-names
        - orcid
        - affiliation

        of which we only need the names.
        """
        # A single author will be read in as a dictionary
        if isinstance(self.authors, dict):
            surname = self.authors["family-names"]
            forename = self.authors["given-names"]
            self.authors = f"{forename} {surname}"
        # Multiple authors will be read in as a list of dictionaries
        elif isinstance(self.authors, list):
            all_authors = []
            for author_info in self.authors:
                if not isinstance(author_info, dict):
                    raise TypeError(
                        "Expected individual author entry to be "
                        "a dictionary, but it was "
                        f"{type(author_info).__name__}: {author_info}"
                    )
                else:
                    surname = author_info["family-names"]
                    forename = author_info["given-names"]
                    all_authors.append(f"{forename} {surname}")
            self.authors = self.author_separator.join(all_authors)
        # Unrecognised read format, abort
        else:
            raise TypeError(
                f"Expected authors to be either dict or list of dicts,"
                f" not {type(self.authors).__name__}"
            )
        return

    def generate_ref_string(self) -> str:
        """
        Generate the citation text from the information provided,
        in the given format.

        This method will be overwritten by the derived class that
        manages the format.
        """
        pass



import inspect
import sys
from dataclasses import dataclass, field
from typing import Any, Dict, List, Set

import requests

from brainglobe_utils.citation.fetch import fetch_from_github, yaml_str_to_dict


@dataclass
class Repository:
    """
    Static class for representing GitHub repositories, in particular
    when needing to fetch CITATION information from them.

    When providing the tool aliases, the tool name will be automatically
    included.
    Likewise, any variations aliases that contain hyphens, underscores, or
    whitespace characters will be automatically given equivalent aliases
    with these characters interchanged.
    Finally, any 'brainglobe' prefixed packages will be given aliases without
    this prefix.
    For example, the name brainglobe-atlasapi would automatically generate
    the following tool aliases:

    - brainglobe atlasapi
    - brainglobe_atlasapi
    - atlasapi

    Parameters
    ----------
    See attributes.

    Attributes
    ----------
    name : str
        The name of the repository.
    tool_aliases: Set[str]
        Names by which the tool the repository provides might be called.
        A single string can be passed if only one alternative alias is to be
        used.
    cff_branch: str, default = "main"
        Branch on which the citation file can be found.
    cff_loc: str, default = "CITATION.cff"
        Location of the citation file on the appropriate branch.
    org: str, default = "brainglobe"
        Organisation or user to which the repository belongs.
    """

    name: str
    tool_aliases: Set[str] = field(default_factory=lambda: set())
    cff_branch: str = "main"
    cff_loc: str = "CITATION.cff"
    org: str = "brainglobe"

    @property
    def url(self) -> str:
        """
        URL to the repository as hosted on GitHub.
        """
        return f"https://github.com/{self.org}/{self.name}"

    def __contains__(self, alias: str) -> bool:
        """
        Syntactic sugar to allow the use of
        if alias in Repository,
        when asking if a Repository is known by the given alias.

        Comparison is case-insensitive for added protection.
        """
        return alias.lower() in [name.lower() for name in self.tool_aliases]

    def __eq__(self, other: "Repository") -> bool:
        """
        Repositories are identical if they point to the same repository.
        """
        return self.url == other.url

    def __hash__(self) -> int:
        """
        Repositories are hashable by their GitHub url.
        """
        return hash(self.url)

    def __lt__(self, other: "Repository") -> bool:
        """
        Repositories are compared by name, alphabetically.
        """
        return self.name < other.name

    def __post_init__(self) -> None:
        """
        Run the following post-init checks:

        - self.tool_aliases is a set, or can be immediately cast to one.
        - The repository actually exists and is reachable.

        And ensure the following behaviour is adhered to:

        - Tools always have an alias identical to their repository name.
        - The underscore, hyphen, and space characters can be used
        interchangeably when referring to tool aliases.
        - A brainglobe-prefixed package automatically gets an alias without
        the prefix present.
        """
        # Cast tool_aliases to a set if provided as some other iterable
        if not isinstance(self.tool_aliases, set) and not isinstance(
            self.tool_aliases, str
        ):
            try:
                self.tool_aliases = set(self.tool_aliases)
            except TypeError as e:
                raise TypeError(
                    "Cannot convert input of type "
                    f"{type(self.tool_aliases).__name__} to a set"
                ) from e
        elif isinstance(self.tool_aliases, str):
            # Strings cast to sets via str -> list -> set, so account for this
            self.tool_aliases = set([self.tool_aliases])

        # Check repository actually exists
        ping_site = requests.get(self.url)
        if not ping_site.ok:
            if ping_site.status_code == 404:
                # Site not found, so repository does not exist
                raise ValueError(
                    f"Repository {self.org}/{self.name} does not exist"
                    " (got 404 response)"
                )
            else:
                raise ValueError(
                    f"Could not reach {self.url} successfully (non 404 status)"
                )

        # These characters are deemed interchangeable for the purposes of
        # supplying and referring to aliases.
        interchangeable = ["-", "_", " "]

        # Can always refer to yourself by repository name
        if self.name not in self.tool_aliases:
            self.tool_aliases.add(self.name)

        # If our name starts with the 'brainglobe' prefix, add another
        # alias that drops this prefix.
        for char in interchangeable:
            if self.name.startswith(f"brainglobe{char}"):
                self.tool_aliases.add(
                    self.name.removeprefix(f"brainglobe{char}")
                )

        # Ensure that hyphens, dashes, and spaces are interchangeable
        # when referring to tool aliases
        # So we don't iterate over a growing set
        original_tool_aliases = self.tool_aliases.copy()
        # For each interchangeable character
        for char in interchangeable:
            replacements = [c for c in interchangeable if c is not char]
            for alias in original_tool_aliases:
                for replacement_char in replacements:
                    proposed_equivalent_alias = alias.replace(
                        char, replacement_char
                    )
                    self.tool_aliases.add(proposed_equivalent_alias)

        return

    def __str__(self) -> str:
        """
        Repositories are represented by their GitHub url.
        """
        return self.url

    def read_citation_info(self) -> Dict[str, Any]:
        """
        Read citation information from the repository into a dictionary.
        """
        cff_response = fetch_from_github(
            self.org, self.name, self.cff_loc, self.cff_branch
        )

        return yaml_str_to_dict(cff_response.text)


def unique_repositories_from_tools(
    *tools: str, report_duplicates: bool = False
) -> Set[Repository]:
    """
    Given a list of tool aliases, return a set of the unique repositories
    that they correspond to.

    Repositories that are referred to by more than one tool in the list
    provided can be reported by flagging the appropriate input.
    These duplicates are then removed from the output.

    Parameters
    ----------
    tools: str
        Tool names or aliases whose repositories should be looked up and
        filtered for duplicates.
    report_duplicates: bool, default = False
        If True, a printout will be sent to the console when a repository
        is referenced more than once in the tool list.

    Returns
    -------
    Set[Repository]
        A set containing the unique repositories that are referenced by the
        tool list.

    Raises
    ------
    RuntimeError
        If one of the tool names provides does not correspond to one of our
        (brainglobe) repositories.
    """
    unique_repos: Set[Repository] = set()

    # Infer the unique repositories from the list of tools
    for tool in tools:
        repo_to_cite: Repository = None
        for repo in all_citable_repositories():
            if tool in repo:
                if repo_to_cite:
                    # We have already found this alias in another repository,
                    # Flag error
                    raise ValueError(
                        f"Multiple repositories match tool {tool}: "
                        f"{repo_to_cite.name}, {repo.name}"
                    )
                else:
                    # This is the first repository that might match the tool
                    repo_to_cite = repo
        if repo_to_cite is None:
            # No repository matches this tool, throw error.
            raise RuntimeError(
                f"No citable repository found for tool {tool}. "
                "If you think this option is missing, please report it: "
                "https://github.com/brainglobe/brainglobe-meta/issues"
            )
        elif (repo_to_cite in unique_repos) and report_duplicates:
            # We already added this repository, so print out a record
            # of the duplication
            print(f"{tool} is already being cited by {repo_to_cite.name}")
        else:
            # Add first occurrence of the repository to the unique list
            unique_repos.add(repo_to_cite)

    return unique_repos


# Static instances for each of our repositories
# that we provide citation information for.
brainglobe_atlasapi = Repository(
    "brainglobe-atlasapi",
    [
        "BrainGlobe AtlasAPI",
        "AtlasAPI",
        "Atlas API",
    ],
)
brainglobe_heatmap = Repository(
    "brainglobe-heatmap",
    [
        "heat-map",
    ],
)
brainglobe_meta = Repository(
    "brainglobe-meta",
    [
        "brainglobe",
        "meta package",
        "toolsuite",
        "tool suite",
    ],
)
brainglobe_napari_io = Repository(
    "brainglobe-napari-io",
    [
        "napari io",
    ],
)
brainglobe_segmentation = Repository(
    "brainglobe-segmentation",
    [
        "brainreg-segment",  # Legacy compatibility
    ],
)
brainglobe_space = Repository(
    "brainglobe-space",
    [
        "bg-space",  # Legacy compatibility
    ],
)
brainglobe_utils = Repository(
    "brainglobe-utils",
    "utilities",
)
brainreg = Repository(
    "brainreg",
    [
        "brainreg-napari",  # Legacy compatibility
    ],
)
brainrender = Repository(
    "brainrender",
)
brainrender_napari = Repository(
    "brainrender-napari",
)
cellfinder = Repository(
    "cellfinder",
    [
        "cellfinder-core",
        "cellfinder-napari",
    ],
)
# MORPHAPI currently broken at time of writing,
# reinstate once fixed and CITATION.cff file added
# morphapi = Repository(
#     "morphapi",
#     [
#         "morph api",
#     ],
# )


def all_citable_repositories() -> List[Repository]:
    """
    Return a list of all citable brainglobe repositories.

    That is, a list of all static repository instances defined in
    this submodule.
    """
    return set(
        obj
        for _, obj in inspect.getmembers(
            sys.modules[__name__], lambda x: isinstance(x, Repository)
        )
    )



from pathlib import Path


def append_to_pathlib_stem(path: Path, string_to_append: str):
    """
    Appends a string to the stem of a pathlib object.

    Parameters
    ----------
    path : pathlib.Path
        Pathlib path.

    string_to_append : str
        String to append to the stem.

    Returns
    -------
    pathlib.Path
        Pathlib object with the string added to the stem.
    """
    return path.parent / (path.stem + string_to_append + path.suffix)



from pathlib import Path
from typing import Any, Dict, Optional, Union

import yaml


def read_yaml_section(yaml_file: Union[str, Path], section: str) -> Any:
    """
    Read section from yaml file.

    Parameters
    ----------
    yaml_file : str or pathlib.Path
        Path of .yml file to read.
    section : str
        Key of yaml section to read.

    Returns
    -------
    The contents of the yaml section.
    """
    yaml_contents = open_yaml(yaml_file)
    contents = yaml_contents[section]
    return contents


def open_yaml(yaml_file: Union[str, Path]) -> Dict[str, Any]:
    """
    Read the contents of a yaml file.

    Parameters
    ----------
    yaml_file : str or pathlib.Path
        Path of .yml file to read.

    Returns
    -------
    dict
        The contents of the yaml file.
    """
    with open(yaml_file) as f:
        yaml_contents = yaml.load(f, Loader=yaml.FullLoader)
    return yaml_contents


def save_yaml(
    yaml_contents: Dict[str, Any],
    output_file: Union[str, Path],
    default_flow_style: Optional[bool] = False,
):
    """
    Save contents to a yaml file.

    Parameters
    ----------
    yaml_contents : dict
        Content to write to yaml file.
    output_file : str or pathlib.Path
        Path to .yml file to write to.
    default_flow_style : bool, optional
        If true, will use block style or flow style depending on whether there
        are nested collections. If false, always uses block style.
    """
    with open(output_file, "w") as outfile:
        yaml.dump(
            yaml_contents, outfile, default_flow_style=default_flow_style
        )



import argparse


def is_even(num):
    """
    Returns True if the non-zero input number is even.

    Parameters
    ----------
    num : int
        Input number.

    Returns
    -------
    bool
        True if number is even, otherwise False.

    Raises
    ------
    NotImplementedError
        If the input number is zero.
    """
    if num == 0:
        raise NotImplementedError(
            "Input number is 0. Evenness of 0 is not defined by this "
            "function."
        )
    if num % 2:
        return False
    else:
        return True


def check_positive_float(value, none_allowed=True):
    """
    Used in argparse to enforce positive floats.
    Source: https://stackoverflow.com/questions/14117415

    Parameters
    ----------
    value : float
        Input value.

    none_allowed : bool, optional
        If False, throw an error for None values.

    Returns
    -------
    float
        Input value, if it's positive.

    Raises
    ------
    argparse.ArgumentTypeError
        If input value is invalid.
    """
    ivalue = value
    if ivalue is not None:
        ivalue = float(ivalue)
        if ivalue < 0:
            raise argparse.ArgumentTypeError(
                "%s is an invalid positive value" % value
            )
    else:
        if not none_allowed:
            raise argparse.ArgumentTypeError("%s is an invalid value." % value)

    return ivalue


def check_positive_int(value, none_allowed=True):
    """
    Used in argparse to enforce positive ints.
    Source: https://stackoverflow.com/questions/14117415

    Parameters
    ----------
    value : int
        Input value.

    none_allowed : bool, optional
        If False, throw an error for None values.

    Returns
    -------
    int
        Input value, if it's positive.

    Raises
    ------
    argparse.ArgumentTypeError
        If input value is invalid.
    """
    ivalue = value
    if ivalue is not None:
        ivalue = int(ivalue)
        if ivalue < 0:
            raise argparse.ArgumentTypeError(
                "%s is an invalid positive value" % value
            )
    else:
        if not none_allowed:
            raise argparse.ArgumentTypeError("%s is an invalid value." % value)

    return ivalue



from configobj import ConfigObj


def get_config_obj(config_path):
    config_path = str(config_path)
    config_obj = ConfigObj(
        config_path,
        encoding="UTF8",
        indent_type="    ",
    )
    return config_obj



import numpy as np


def get_bins(image_size, bin_sizes):
    """
    Given an image size, and bin size, return a list of the bin boundaries.

    Parameters
    ----------
    image_size : tuple of int or list of int
        Size of the final image
    bin_sizes : tuple of int or list of int
        Bin sizes corresponding to the dimensions of "image_size"

    Returns
    -------
    list of np.ndarray
        List of arrays of bin boundaries
    """
    bins = []
    for dim in range(0, len(image_size)):
        bins.append(np.arange(0, image_size[dim] + 1, bin_sizes[dim]))
    return bins



"""
Parts based on https://github.com/SainsburyWellcomeCentre/cell_count_analysis
by Charly Rousseau (https://github.com/crousseau).
"""

from dataclasses import dataclass
from pathlib import Path
from typing import List, Set, Union

import numpy as np
import pandas as pd
from brainglobe_atlasapi import BrainGlobeAtlas

from brainglobe_utils.general.system import ensure_directory_exists
from brainglobe_utils.pandas.misc import safe_pandas_concat, sanitise_df


@dataclass
class Point:
    """
    A class to represent a point in both raw and atlas coordinate spaces,
    along with associated anatomical information.

    Attributes
    ----------
    raw_coordinate : np.ndarray
        A numpy array representing the raw coordinates of the point
        in the original image space.
    atlas_coordinate : np.ndarray
        A numpy array representing the coordinates of the point in atlas space.
    structure : str
        The name of the atlas structure associated with the point.
    structure_id : int
        The numerical ID of the anatomical structure associated with the point.
    hemisphere : str
        The hemisphere ('left' or 'right') in which the point is located.

    """

    raw_coordinate: np.ndarray
    atlas_coordinate: np.ndarray
    structure: str
    structure_id: int
    hemisphere: str


def calculate_densities(
    counts: pd.DataFrame, volume_csv_path: Union[str, Path]
) -> pd.DataFrame:
    """
    Use region volumes from registration to calculate cell densities.

    Parameters
    ----------
    counts : pd.DataFrame
        Dataframe with cell counts.
    volume_csv_path : Union[str, Path]
        Path of the CSV file containing the volumes of each brain region.

    Returns
    -------
    pd.DataFrame
        A dataframe containing the original cell counts and the calculated cell
        densities per mm³ for each brain region.

    """

    volumes = pd.read_csv(volume_csv_path, sep=",", header=0, quotechar='"')
    df = pd.merge(counts, volumes, on="structure_name", how="outer")
    df = df.fillna(0)
    df["left_cells_per_mm3"] = df.left_cell_count / df.left_volume_mm3
    df["right_cells_per_mm3"] = df.right_cell_count / df.right_volume_mm3
    return df


def combine_df_hemispheres(df: pd.DataFrame) -> pd.DataFrame:
    """
    Combine left and right hemisphere data onto a single row

    Parameters
    ----------
    df : pd.DataFrame
        A pandas DataFrame with separate rows for left and
        right hemisphere data.

    Returns
    -------
    pd.DataFrame
        A DataFrame with combined hemisphere data. Each row
        represents the combined data of left and right hemispheres for
        each brain region.

    """
    left = df[df["hemisphere"] == "left"]
    right = df[df["hemisphere"] == "right"]
    left = left.drop(["hemisphere"], axis=1)
    right = right.drop(["hemisphere"], axis=1)
    left.rename(columns={"cell_count": "left_cell_count"}, inplace=True)
    right.rename(columns={"cell_count": "right_cell_count"}, inplace=True)
    both = pd.merge(left, right, on="structure_name", how="outer")
    both = both.fillna(0)
    both["total_cells"] = both.left_cell_count + both.right_cell_count
    both = both.sort_values("total_cells", ascending=False)
    return both


def create_all_cell_csv(
    points: List[Point], output_filename: Union[str, Path]
) -> None:
    """
    Create a CSV file with cell data from a list of Point objects.

    This function takes a list of Point objects, each representing cell
    coordinates and brain region and converts this into a pandas DataFrame.
    The DataFrame is then saved to a CSV file at the specified filename.

    Parameters
    ----------
    points : List[Point]
        A list of Point objects, each containing cell data such as
        raw and atlas coordinates,
        structure name, and hemisphere information.
    output_filename : Union[str, Path]
        The filename (including path) where the CSV file will be saved.
        Can be a string or a Path object.

    Returns
    -------
    None
    """

    ensure_directory_exists(Path(output_filename).parent)
    df = pd.DataFrame(
        columns=(
            "coordinate_raw_axis_0",
            "coordinate_raw_axis_1",
            "coordinate_raw_axis_2",
            "coordinate_atlas_axis_0",
            "coordinate_atlas_axis_1",
            "coordinate_atlas_axis_2",
            "structure_name",
            "hemisphere",
        )
    )

    temp_matrix = [[] for i in range(len(points))]
    for i, point in enumerate(points):
        temp_matrix[i].append(point.raw_coordinate[0])
        temp_matrix[i].append(point.raw_coordinate[1])
        temp_matrix[i].append(point.raw_coordinate[2])
        temp_matrix[i].append(point.atlas_coordinate[0])
        temp_matrix[i].append(point.atlas_coordinate[1])
        temp_matrix[i].append(point.atlas_coordinate[2])
        temp_matrix[i].append(point.structure)
        temp_matrix[i].append(point.hemisphere)

    df = pd.DataFrame(temp_matrix, columns=df.columns, index=None)
    df.to_csv(output_filename, index=False)


def count_points_per_brain_region(
    points: List[Point],
    structures_with_points: Set[str],
    brainreg_volume_csv_path: Union[str, Path],
    output_filename: Union[str, Path],
) -> None:
    """
    Count the number of points per brain region.

    Parameters
    ----------
    points : List[Point]
        A list of Point objects.
    structures_with_points : Set[str]
        A set of strings representing the names of all atlas structures
        represented
    brainreg_volume_csv_path : Union[str, Path]
        The path to the CSV file containing volume information from the
        brainreg registration.
    output_filename : Union[str, Path]
        The path where the summary of points by atlas region will be saved.

    Returns
    -------
    None
    """

    structures_with_points = list(structures_with_points)

    point_numbers = pd.DataFrame(
        columns=("structure_name", "hemisphere", "cell_count")
    )
    for structure in structures_with_points:
        for hemisphere in ("left", "right"):
            n_points = len(
                [
                    point
                    for point in points
                    if point.structure == structure
                    and point.hemisphere == hemisphere
                ]
            )
            if n_points:
                point_numbers = safe_pandas_concat(
                    point_numbers,
                    pd.DataFrame(
                        data=[[structure, hemisphere, n_points]],
                        columns=[
                            "structure_name",
                            "hemisphere",
                            "cell_count",
                        ],
                    ),
                )
    sorted_point_numbers = point_numbers.sort_values(
        by=["cell_count"], ascending=False
    )

    combined_hemispheres = combine_df_hemispheres(sorted_point_numbers)
    df = calculate_densities(combined_hemispheres, brainreg_volume_csv_path)
    df = sanitise_df(df)

    df.to_csv(output_filename, index=False)


def summarise_points_by_atlas_region(
    points_in_raw_data_space: np.ndarray,
    points_in_atlas_space: np.ndarray,
    atlas: BrainGlobeAtlas,
    brainreg_volume_csv_path: Union[str, Path],
    points_list_output_filename: Union[str, Path],
    summary_filename: Union[str, Path],
) -> None:
    """
    Summarise points data by atlas region.

    This function takes points in both raw data space and atlas space,
    and generates a summary of these points based on the BrainGlobe atlas
    region they are found in.
    The summary is saved to a CSV file.

    Parameters
    ----------
    points_in_raw_data_space : np.ndarray
        A numpy array representing points in the raw data space.
        The shape of the array should be (n_points, 3).
    points_in_atlas_space : np.ndarray
        A numpy array representing points in the atlas space.
        The shape of the array should be (n_points, 3).
    atlas : BrainGlobeAtlas
        The BrainGlobe atlas object used for the analysis
    brainreg_volume_csv_path : Union[str, Path]
        The path to the CSV file containing volume information from the
        brainreg registration.
    points_list_output_filename : Union[str, Path]
        The csv path where the detailed points list will be saved.
    summary_filename : Union[str, Path]
        The csv path where the summary of points by atlas region will be saved.

    Returns
    -------
    None
    """

    points = []
    structures_with_points = set()
    for idx, point in enumerate(points_in_atlas_space):
        try:
            structure_id = atlas.structure_from_coords(point)
            structure = atlas.structures[structure_id]["name"]
            hemisphere = atlas.hemisphere_from_coords(point, as_string=True)
            points.append(
                Point(
                    points_in_raw_data_space[idx],
                    point,
                    structure,
                    structure_id,
                    hemisphere,
                )
            )
            structures_with_points.add(structure)
        except Exception:
            continue

    create_all_cell_csv(points, points_list_output_filename)

    count_points_per_brain_region(
        points,
        structures_with_points,
        brainreg_volume_csv_path,
        summary_filename,
    )



import logging
import os
from typing import List, Optional, Tuple

import brainglobe_space as bgs
import numpy as np
import pandas as pd
import tifffile
from brainglobe_atlasapi import BrainGlobeAtlas

from brainglobe_utils.image_io import get_size_image_from_file_paths


def transform_points_from_downsampled_to_atlas_space(
    downsampled_points: np.ndarray,
    atlas: BrainGlobeAtlas,
    deformation_field_paths: List[os.PathLike],
    output_filename: Optional[os.PathLike] = None,
    warn_out_of_bounds: bool = True,
) -> Tuple[np.ndarray, List]:
    """
    Transform points from the "downsampled" space in brainreg (i.e. raw data
    transformed to the same orientation and scale as the atlas) to the atlas
    coordinate space.


    Parameters
    ----------
    downsampled_points :
        Points already downsampled to the atlas resolution.
    atlas :
        Target BrainGlobe atlas.
    deformation_field_paths :
        File paths to the deformation fields as generated by brainreg.
    output_filename :
        File to save transformed points to. Points are saved as a HDF file
        using pandas.
    warn_out_of_bounds :
        Whether to warn if points fall outside of the atlas space.

    Returns
    -------
    transformed_points
        Points transformed to the atlas space.

    points_out_of_bounds
        Points that fall outside of the transformed atlas space.
    """

    field_scales = [int(1000 / resolution) for resolution in atlas.resolution]
    points: List[List] = [[], [], []]
    points_out_of_bounds = []
    for axis, deformation_field_path in enumerate(deformation_field_paths):
        deformation_field = tifffile.imread(deformation_field_path)
        for point in downsampled_points:
            try:
                point = [int(round(p)) for p in point]
                points[axis].append(
                    int(
                        round(
                            field_scales[axis]
                            * deformation_field[point[0], point[1], point[2]]
                        )
                    )
                )
            except IndexError:
                if point not in points_out_of_bounds:
                    points_out_of_bounds.append(point)
                    if warn_out_of_bounds:
                        logging.info(
                            f"Ignoring point: {point} "
                            f"as it falls outside the atlas."
                        )

    transformed_points = np.array(points).T

    if output_filename is not None:
        df = pd.DataFrame(transformed_points)
        df.to_hdf(output_filename, key="df", mode="w")

    return transformed_points, points_out_of_bounds


def get_anatomical_space_from_image_planes(
    image_plane: np.ndarray, orientation: str, voxel_sizes: List[float]
) -> bgs.AnatomicalSpace:
    """
    Determine the brainglobe anatomical space from an image plane.

    Parameters
    ----------
    image_plane : np.ndarray
        A numpy-like array representing a single image plane from a 3D image.
    orientation : str
        The orientation of the image following the brainglobe-space
        three-letter convention (e.g., 'asr', 'psl').
    voxel_sizes : List[float]
        A list of floats representing the voxel sizes (e.g., [5, 2, 2]).

    Returns
    -------
    bgs.AnatomicalSpace
        An object representing the anatomical space.

    """

    shape = tuple(get_size_image_from_file_paths(image_plane).values())

    space = bgs.AnatomicalSpace(
        orientation,
        shape=(shape[2], shape[1], shape[0]),
        resolution=[float(i) for i in voxel_sizes],
    )
    return space


def transform_points_from_raw_to_downsampled_space(
    points: np.ndarray,
    source_image_plane: np.ndarray,
    orientation: str,
    voxel_sizes: List[float],
    downsampled_space: bgs.AnatomicalSpace,
    output_filename: Optional[os.PathLike] = None,
) -> np.ndarray:
    """
    Transform points from raw data space to the "downsampled" space as
    defined in brainreg (same orientation and voxel size as the atlas)

    Parameters
    ----------
    points : np.ndarray
        Points in the original space.
    source_image_plane : np.ndarray
        A numpy-like array representing a single image.
    orientation : str
        The orientation of the image following the brainglobe-space
    three letter convention (e.g. 'asr', 'psl')
    voxel_sizes : List[float]
        A list of floats representing the voxel sizes (e.g. [5, 2, 2])
    downsampled_space : bgs.AnatomicalSpace
        Target anatomical space.
    output_filename : Optional[os.PathLike], optional
        File to save downsampled points to. Points are saved as a HDF file
        using pandas.

    Returns
    -------
    np.ndarray
        Points transformed to the downsampled space.

    """
    source_space = get_anatomical_space_from_image_planes(
        source_image_plane, orientation, voxel_sizes
    )
    points = source_space.map_points_to(downsampled_space, points)
    if output_filename is not None:
        df = pd.DataFrame(points)
        df.to_hdf(output_filename, key="df", mode="w")
    return points


def transform_points_to_atlas_space(
    points: np.ndarray,
    source_image_plane: np.ndarray,
    orientation: str,
    voxel_sizes: List[float],
    downsampled_space: bgs.AnatomicalSpace,
    atlas: BrainGlobeAtlas,
    deformation_field_paths: List[os.PathLike],
    downsampled_points_path: Optional[os.PathLike] = None,
    atlas_points_path: Optional[os.PathLike] = None,
) -> np.ndarray:
    """
    Transform points from raw data to an atlas space.

    The points are first downsampled to the atlas resolution, and then
    transformed to the atlas space.

    Parameters
    ----------
    points : np.ndarray
        Points in the original space.
    source_image_plane : np.ndarray
        A numpy-like array representing a single image.
    orientation : str
        The orientation of the image following the brainglobe-space
    three letter convention (e.g. 'asr', 'psl')
    voxel_sizes : List[float]
        A list of floats representing the voxel sizes (e.g. [5, 2, 2])
    downsampled_space : bgs.AnatomicalSpace
        The BrainGlobe anatomical space representing the downsampled space
    atlas : BrainGlobeAtlas
        The BrainGlobe atlas object used for registration
    deformation_field_paths : List[os.PathLike]
        File paths to the deformation fields as generated by brainreg.
    downsampled_points_path : Optional[os.PathLike], optional
        The file path where the downsampled points
        should be saved (default is None).
    atlas_points_path : Optional[os.PathLike], optional
        The file path where the atlas-transformed points
        should be saved (default is None).

    Returns
    -------
    np.ndarray
        An array of points transformed into the atlas space.

    """
    downsampled_points = transform_points_from_raw_to_downsampled_space(
        points,
        source_image_plane,
        orientation,
        voxel_sizes,
        downsampled_space,
        output_filename=downsampled_points_path,
    )
    points = transform_points_from_downsampled_to_atlas_space(
        downsampled_points,
        atlas,
        deformation_field_paths,
        output_filename=atlas_points_path,
    )

    return points



"""
Cell position I/O

Based on https://github.com/SainsburyWellcomeCentre/niftynet_cell_count by
Christian Niedworok (https://github.com/cniedwor).
"""

import logging
import os
from pathlib import Path
from typing import List, Optional, Union
from xml.dom import minidom
from xml.etree import ElementTree
from xml.etree.ElementTree import Element as EtElement

import pandas as pd
import yaml

from brainglobe_utils.cells.cells import (
    Cell,
    MissingCellsError,
    UntypedCell,
    pos_from_file_name,
)
from brainglobe_utils.general.system import replace_extension


def get_cells(
    cells_file_path: str,
    cells_only: bool = False,
    cell_type: Optional[int] = None,
):
    """
    Read cells from a file or directory.

    Parameters
    ----------
    cells_file_path : str
        Path to cells file to read. Can be .xml, .yml, or a directory.
    cells_only : bool, optional
        Only relevant for .xml files. If True, will only read Cells with
        type Cell.CELL i.e. exclude all Cell.ARTIFACT, Cell.UNKNOWN, and
        Cell.NO_CELL.
    cell_type : int, optional
        Only relevant for directories. This determines the type to assign
        to all read cells. Refer to the Cell documentation for
        a full explanation of cell_type options.

    Returns
    -------
    list of Cell
        A list of the cells contained in the file or directory.

    See Also
    --------
    Cell : For full description of cell_type parameter.
    """
    if cells_file_path.endswith(".xml"):
        return get_cells_xml(cells_file_path, cells_only=cells_only)
    elif cells_file_path.endswith(".yml"):
        return get_cells_yml(cells_file_path, ignore_type=True)
    elif os.path.isdir(cells_file_path):
        try:
            return get_cells_dir(cells_file_path, cell_type=cell_type)
        except IndexError:
            # if a directory is given, but it contains
            # files that can't be read. Usually if the user gives the wrong
            # directory as input to `cellfinder_gen_cubes`
            raise_cell_read_error(cells_file_path)
    else:
        raise_cell_read_error(cells_file_path)


def raise_cell_read_error(cells_file_path):
    """Raise a NotImplementedError, with an informative message including the
    cells file path"""
    logging.error(
        "File format of: {} is not supported or contains errors. Please "
        "supply an xml file, or a directory of files with positions in the "
        "filenames."
        "".format(cells_file_path)
    )
    raise NotImplementedError(
        "File format of: {} is not supported or contains errors. Please "
        "supply an xml file, or a directory of files with positions in the "
        "filenames."
        "".format(cells_file_path)
    )


def get_cells_xml(
    xml_file_path: Union[str, Path], cells_only: Optional[bool] = False
):
    """
    Read cells from an xml file.

    Parameters
    ----------
    xml_file_path : str or pathlib.Path
        Path to xml file to read from.
    cells_only : bool, optional
        Whether to only read Cells with type Cell.CELL i.e. exclude all
        Cell.ARTIFACT, Cell.UNKNOWN, and Cell.NO_CELL.

    Returns
    -------
    list of Cell
        A list of the cells contained in the file.
    """
    with open(xml_file_path, "r") as xml_file:
        root = ElementTree.parse(xml_file).getroot()
        cells = []
        for type_marker in root.find("Marker_Data").findall("Marker_Type"):
            cell_type = int(type_marker.find("Type").text)
            for cell_element in type_marker.findall("Marker"):
                cells.append(Cell(cell_element, cell_type))
        if not cells:
            raise MissingCellsError(
                "No cells found in file {}".format(xml_file_path)
            )
    if cells_only:
        cells = [c for c in cells if c.is_cell()]
    return cells


def get_cells_yml(
    cells_file_path: Union[str, Path],
    ignore_type: Optional[bool] = False,
    marker: Optional[str] = "markers",
):
    """
    Read cells from a yml file.

    Parameters
    ----------
    cells_file_path : str or pathlib.Path
        Path to yml file to read from.
    ignore_type : bool, optional
        Whether to ignore the type of cells - all will be assigned type
        Cell.UNKNOWN. Currently only True is supported.
    marker : str, optional
        Yaml key under which cells information is stored.

    Returns
    -------
    list of Cell
        A list of the cells contained in the file.
    """
    if not ignore_type:
        raise NotImplementedError(
            "Parsing cell types is not yet implemented for YAML files. "
            "Currently the only option is to merge them. Please try again with"
            " 'ignore_type=True'."
        )
    else:
        with open(cells_file_path, "r") as yml_file:
            data = yaml.safe_load(yml_file)
        cells = []
        for cell_type in list(data.keys()):
            type_dict = data[cell_type]
            if marker in type_dict.keys():
                for cell in type_dict[marker]:
                    cells.append(Cell(cell, Cell.UNKNOWN))
    return cells


def get_cells_dir(
    cells_file_path: Union[str, Path], cell_type: Optional[bool] = None
):
    """
    Read cells from a directory. Cells will be created based on the filenames
    of files in the directory, one cell per file.

    Parameters
    ----------
    cells_file_path : str or pathlib.Path
        Path to directory containing Cells. Each file in the directory will
        create one Cell, based on its filename. This filename must contain the
        cell's x, y and z position E.g. 'Cell_z358_y564_x4056'
    cell_type : int or str or None, optional
        What type to assign all cells. Refer to the Cell documentation for
        a full explanation of cell_type options.

    Returns
    -------
    list of Cell
        A list of the cells contained in the directory.

    See Also
    --------
    Cell : For full description of cell_type parameter.
    """
    cells = []
    for file in os.listdir(cells_file_path):
        # ignore hidden files
        if not file.startswith("."):
            cells.append(Cell(file, cell_type))
    return cells


def save_cells(
    cells,
    xml_file_path,
    save_csv=False,
    indentation_str="  ",
    artifact_keep=True,
):
    """
    Save cells to a file.

    Parameters
    ----------
    cells : list of Cell
        Cells to save to file.
    xml_file_path : str
        File path of xml file to save cells to.
    save_csv : bool, optional
        If True, will save cells to a csv file (rather than xml). This will use
        the given xml_file_path with the .xml extension replaced with .csv.
    indentation_str : str, optional
        String to use for indent in xml file.
    artifact_keep : bool, optional
        Whether to keep cells with type Cell.ARTIFACT in the xml file. If True,
        they are kept but their type is changed to Cell.UNKNOWN. If False, they
        are removed.
    """
    # Assume always save xml file, and maybe save other formats
    cells_to_xml(
        cells,
        xml_file_path,
        indentation_str=indentation_str,
        artifact_keep=artifact_keep,
    )

    if save_csv:
        csv_file_path = replace_extension(xml_file_path, "csv")
        cells_to_csv(cells, csv_file_path)


def cells_to_xml(
    cells: List[Cell],
    xml_file_path: Union[str, Path],
    indentation_str: Optional[str] = "  ",
    artifact_keep: Optional[bool] = True,
):
    """
    Save cells to an xml file.

    Parameters
    ----------
    cells : list of Cell
        Cells to save to file.
    xml_file_path : str or pathlib.Path
        Xml file path to write cells to.
    indentation_str : str, optional
        String to use for indent in xml file.
    artifact_keep : bool, optional
        Whether to keep cells with type Cell.ARTIFACT. If True, they
        are kept but their type is changed to Cell.UNKNOWN. If False, they are
        removed.
    """
    xml_data = make_xml(cells, indentation_str, artifact_keep=artifact_keep)
    with open(xml_file_path, "w") as xml_file:
        xml_file.write(str(xml_data, "UTF-8"))


def cells_xml_to_df(xml_file_path):
    """Read cells from xml file and convert to dataframe"""
    cells = get_cells(xml_file_path)
    return cells_to_dataframe(cells)


def cells_to_dataframe(cells: List[Cell]) -> pd.DataFrame:
    return pd.DataFrame([c.to_dict() for c in cells])


def cells_to_csv(cells: List[Cell], csv_file_path: Union[str, Path]):
    """Save cells to csv file"""
    df = cells_to_dataframe(cells)
    df.to_csv(csv_file_path)


def make_xml(cell_list, indentation_str, artifact_keep=True):
    """
    Convert a list of cells to xml.

    Parameters
    ----------
    cell_list : list of Cell
        Cells to convert to xml.
    indentation_str : str
        String to use for indent in xml file.
    artifact_keep : bool, optional
        Whether to keep cells with type Cell.ARTIFACT. If True, they
        are kept but their type is changed to Cell.UNKNOWN. If False, they are
        removed.

    Returns
    -------
    bytes
        Cell list as xml, ready to be saved into a file.
    """
    root = EtElement("CellCounter_Marker_File")
    image_properties = EtElement("Image_Properties")
    file_name = EtElement("Image_Filename")
    file_name.text = "placeholder.tif"
    image_properties.append(file_name)
    root.append(image_properties)

    marker_data = EtElement("Marker_Data")
    current_type = EtElement("Current_Type")
    current_type.text = str(1)  # TODO: check
    marker_data.append(current_type)

    cell_dict = make_type_dict(cell_list)

    # if artifacts exist, do something with them (convert/delete)
    if Cell.ARTIFACT in cell_dict:
        cell_dict = deal_with_artifacts(cell_dict, artifact_keep=artifact_keep)

    for cell_type, cells in cell_dict.items():
        type_el = EtElement("Type")
        type_el.text = str(cell_type)
        mt = EtElement("Marker_Type")
        mt.append(type_el)
        for cell in cells:
            mt.append(cell.to_xml_element())
        marker_data.append(mt)
    root.append(marker_data)

    return pretty_xml(root, indentation_str)


def deal_with_artifacts(cell_dict, artifact_keep=True):
    """Deal with cells with type Cell.ARTIFACT in a cell dict. If artifact_keep
    is True, then convert all Cell.ARTIFACT to Cell.UNKNOWN. Otherwise, remove
    all Cell.ARTIFACT"""
    # What do we want to do with the artifacts (if they exist)?
    # Might want to keep them for training
    if artifact_keep:
        logging.debug("Keeping artifacts")
        # Change their type
        for idx, artifact in enumerate(cell_dict[Cell.ARTIFACT]):
            cell_dict[Cell.ARTIFACT][idx].type = Cell.UNKNOWN
        # Add them to "cell_type = UNKNOWN" list
        cell_dict[Cell.UNKNOWN].extend(cell_dict[Cell.ARTIFACT])
    else:
        logging.debug("Removing artifacts")
    del cell_dict[Cell.ARTIFACT]  # outside if, needs to be run regardless
    return cell_dict


def make_type_dict(cell_list):
    """Convert a list of Cells to a dictionary with keys of cell type and
    values of the list of corresponding cells."""
    types = sorted(set([cell.type for cell in cell_list]))
    return {
        cell_type: [cell for cell in cell_list if cell.type == cell_type]
        for cell_type in types
    }


def pretty_xml(elem, indentation_str="  "):
    """Convert xml element to pretty version, using given indentation string"""
    ugly_xml = ElementTree.tostring(elem, "utf-8")
    md_parsed = minidom.parseString(ugly_xml)
    return md_parsed.toprettyxml(indent=indentation_str, encoding="UTF-8")


def find_relevant_tiffs(tiffs, cell_def):
    """
    Find tiffs that match those read from cell_def.

    Parameters
    ----------
    tiffs : list of str
        List of paths to tiff files. Each tiff filename must contain the cell's
        x, y and z position.
    cell_def : str
        Path to cells to read. Can be an .xml file, .yml file, or a directory.
        If a directory is passed, each file in the directory will
        create one Cell, based on its filename. Each filename must contain the
        cell's x, y and z position.

    Returns
    -------
    list of str
        Filtered list of paths to tiff files, only including those that match
        cells read from cell_def.
    """
    cells = [UntypedCell(tiff) for tiff in tiffs]
    if os.path.isdir(cell_def):
        relevant_cells = set(
            [UntypedCell(pos_from_file_name(f)) for f in os.listdir(cell_def)]
        )
    else:
        relevant_cells = set(
            [UntypedCell.from_cell(cell) for cell in get_cells(cell_def)]
        )
    return [
        tiffs[pos]
        for (pos, cell) in enumerate(cells)
        if cell in relevant_cells
    ]



import inspect
import sys
from string import ascii_letters, digits
from typing import Any, ClassVar, Dict

from brainglobe_utils.citation.format import Format


class BibTexEntry(Format):
    """
    An abstract base class for generating BibTex entries from
    CITATION.cff yaml-content.

    Constructed by passing a dict containing the yaml-processed
    content of the CITATION.cff file to the constructor.
    Required fields are checked for on instantiation.
    Missing optional fields will be set to None if not provided.

    Parameters
    ----------
    information : Dict[str, str]
        yaml-processed content of CITATION.cff.
    cite_key : str, default = "BrainGlobeReference"
        Citation key to add to the BibTex reference that is generated.
    warn_on_not_used: bool, default = False
        If True, warn the user about the fields in the information
        input that are not required nor optional, so are ignored.

    Attributes
    ----------
    cite_key : str
        The citation key that appears in the BibTex reference
    """

    author_separator: ClassVar[str] = " and "
    cite_key: str

    @classmethod
    def validate_citation_key(cls, key: str) -> bool:
        """
        Return True if the citation key provided is valid for use,
        otherwise return False.

        Valid citation keys may contain;
        - Alphanumeric characters,
        - Underscores (_),
        - Single dashes (-),
        - Semicolons (:),

        and no others.
        """
        bad_characters = key.lower()
        for valid_character in ascii_letters + digits + "_-:":
            bad_characters = bad_characters.replace(valid_character, "")

        if bad_characters:
            # Some characters in the string provided are not permitted,
            # as we have removed all of the permitted characters.
            return False
        return True

    def __init__(
        self,
        information: Dict[str, Any],
        cite_key: str = "BrainGlobeReference",
        warn_on_not_used: bool = False,
    ) -> None:
        """ """
        # So we don't delete information that we may need in other places
        # (C++ memory sharing rights plz Python)
        information = information.copy()

        # Add the citation key if provided,
        # or use the default otherwise
        if self.validate_citation_key(cite_key):
            self.cite_key = cite_key
        else:
            raise ValueError(
                f"Citation key {cite_key} is not valid."
                " Citation keys may only be composed of"
                "alphanumeric characters, digits, '_', '-', and ':'"
            )

        # If type information is available, ensure that we are reading
        # into the correct reference type!
        if "type" in information.keys():
            assert information["type"] == self.entry_type(), (
                "Attempting to read reference of type"
                f" {information['type']} into {self.entry_type()}"
            )
            # Remove type field from information dict,
            # so we don't try to assign it to a field later.
            information.pop("type")

        super().__init__(information, warn_on_not_used=warn_on_not_used)

        return

    def generate_ref_string(self) -> str:
        """
        Generate a string that encodes the reference, in preparation for
        writing to an output format.
        """
        output_string = f"@{self.entry_type()}{{{self.cite_key},\n"

        # Tracks current indentation level
        indent_level = 1

        # Required fields are guaranteed to exist
        for req_field in self.required:
            output_string += (
                self.indent_character * indent_level
                + f'{req_field} = "{getattr(self, req_field)}",\n'
            )

        # Optional fields may be skipped
        for opt_field in self.optional:
            if getattr(self, opt_field):
                output_string += (
                    self.indent_character * indent_level
                    + f'{opt_field} = "{getattr(self, opt_field)}",\n'
                )

        indent_level -= 1
        output_string += "}"

        return output_string


class Article(BibTexEntry):
    """
    Derived class for writing BibTex references to articles.
    """

    required = ["authors", "title", "journal", "year"]
    optional = [
        "volume",
        "number",
        "pages",
        "month",
        "note",
        "doi",
        "issn",
        "zblnumber",
        "eprint",
    ]


class Software(BibTexEntry):
    """
    Derived class for writing BibTex references to software.
    """

    required = ["authors", "title", "url", "year"]
    optional = [
        "abstract",
        "date",
        "doi",
        "eprint",
        "eprintclass",
        "eprinttype",
        "file",
        "hal_id",
        "hal_version",
        "institution",
        "license",
        "month",
        "note",
        "organization",
        "publisher",
        "related",
        "relatedtype",
        "relatedstring",
        "repository",
        "swhid",
        "urldate",
        "version",
    ]


def supported_bibtex_entry_types() -> Dict[str, BibTexEntry]:
    """
    Create a dict of all the classes in this module that can be used
    to write a bibtex reference of a particular entry type.

    keys are the entry type as it will appear in the .tex entry.
    values are the corresponding derived class to use when writing a
    reference of that type.

    Returns
    -------
    Dict[str, @_BibTexEntry]
        Dict of classes derived from BibTexEntry that can handle entry
        types, indexed by the entry type they support.
    """
    this_module = sys.modules[__name__]

    dict_of_formats: Dict[str, BibTexEntry] = {
        cls.entry_type(): cls
        for cls in (
            getattr(this_module, name)
            for name, _ in inspect.getmembers(this_module, inspect.isclass)
            if name != "BibTexEntry" and name != "Format"
        )
        if issubclass(cls, BibTexEntry)
    }

    return dict_of_formats



import warnings
from pathlib import Path

import nrrd
import numpy as np
import tifffile

with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    import nibabel as nib


def to_nii(img, dest_path, scale=None, affine_transform=None):
    # TODO: see if we want also real units scale

    """
    Write the brain volume to disk as a nifty image.

    Parameters
    ----------
    img : nifty image object or np.ndarray
        A nifty image object or numpy array representing a brain.

    dest_path : str or pathlib.Path
        The file path where to save the brain.

    scale : tuple of floats, optional
        A tuple of floats to indicate the 'zooms' of the nifty image.

    affine_transform : np.ndarray, optional
        A 4x4 matrix indicating the transform to save in the metadata of the
        image. Required only if not nibabel input.
    """
    dest_path = Path(dest_path)
    if affine_transform is None:
        affine_transform = np.eye(4)
    if not isinstance(img, nib.Nifti1Image):
        img = nib.Nifti1Image(img, affine_transform)
    if scale is not None:
        img.header.set_zooms(scale)
    nib.save(img, dest_path)


def to_tiff(img_volume, dest_path, photometric="minisblack"):
    """
    Save the image volume (numpy array) to a tiff stack.

    Parameters
    ----------
    img_volume : np.ndarray
        The image to be saved.

    dest_path : str or pathlib.Path
        The file path where to save the tiff stack.

    photometric: str
        Color space of image (to pass to tifffile.imwrite)
        Use 'minisblack' (default) for grayscale and 'rgb' for rgb
    """
    dest_path = Path(dest_path)
    tifffile.imwrite(dest_path, img_volume, photometric=photometric)


def to_tiffs(img_volume, path_prefix, path_suffix="", extension=".tif"):
    """
    Save the image volume (numpy array) as a sequence of tiff planes.
    Each plane will have a filepath of the following format:
    {path_prefix}_{zeroPaddedIndex}{path_suffix}{extension}

    Parameters
    ----------
    img_volume : np.ndarray
        The image to be saved.

    path_prefix : str or pathlib.Path
        The prefix for each plane.

    path_suffix : str, optional
        The suffix for each plane.

    extension : str, optional
        The file extension for each plane.
    """
    if isinstance(path_prefix, Path):
        path_prefix = str(path_prefix.resolve())

    z_size = img_volume.shape[0]
    pad_width = int(round(z_size / 10)) + 1
    for i in range(z_size):
        img = img_volume[i, :, :]
        dest_path = (
            f"{path_prefix}_{str(i).zfill(pad_width)}{path_suffix}{extension}"
        )
        tifffile.imwrite(dest_path, img)


def to_nrrd(img_volume, dest_path):
    """
    Save the image volume (numpy array) as nrrd.

    Parameters
    ----------
    img_volume : np.ndarray
        The image to be saved.

    dest_path : str or pathlib.Path
        The file path where to save the nrrd image.
    """
    if isinstance(dest_path, Path):
        dest_path = str(dest_path.resolve())

    nrrd.write(dest_path, img_volume)


def save_any(img_volume, dest_path):
    """
    Save the image volume (numpy array) to the given file path, using the save
    function matching its file extension.

    Parameters
    ----------
    img_volume : np.ndarray
        The image to be saved.

    dest_path : str or pathlib.Path
        The file path to save the image to.
        Supports directories (will save a sequence of tiffs), .tif, .tiff,
        .nrrd and .nii.
    """
    dest_path = Path(dest_path)

    if dest_path.is_dir():
        to_tiffs(img_volume, dest_path / "image")

    elif dest_path.suffix in [".tif", ".tiff"]:
        to_tiff(img_volume, dest_path)

    elif dest_path.suffix == ".nrrd":
        to_nrrd(img_volume, dest_path)

    elif dest_path.suffix == ".nii":
        to_nii(img_volume, dest_path)

    else:
        raise NotImplementedError(
            f"Could not guess data type for path {dest_path}"
        )



import numpy as np


def scale_to_16_bits(img):
    """
    Normalise the input image to the full 0-2^16 bit depth.

    Parameters
    ----------
    img : np.ndarray
        The input image

    Returns
    -------
    np.ndarray
        The normalised image
    """
    normalised = img / img.max()
    return normalised * (2**16 - 1)


def scale_and_convert_to_16_bits(img):
    """
    Normalise the input image to the full 0-2^16 bit depth, and return as
    type: "np.uint16".

    Parameters
    ----------
    img : np.ndarray
        The input image

    Returns
    -------
    np.ndarray
        The normalised, 16 bit image
    """
    img = scale_to_16_bits(img)
    return img.astype(np.uint16, copy=False)
